/******************************************************************************
 * Copyright (c) 2013-2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/
#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include "ocl_util.h"

using namespace cl;
using namespace std;

const char * kernelStr = "kernel void Null(int a) {}";

static double clock_diff (struct timespec *t1, struct timespec *t2)
           { return t2->tv_sec - t1->tv_sec + (t2->tv_nsec - t1->tv_nsec) / 1e9; }


int main(int argc, char *argv[])
{
   struct timespec tp_start, tp_end;

   try 
   {
     Context context(CL_DEVICE_TYPE_ACCELERATOR);
     std::vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();

     Program::Sources    source(1, std::make_pair(kernelStr,strlen(kernelStr)));
     Program             program = Program(context, source);
     program.build(devices); 

     Kernel kernel(program, "Null");
     kernel.setArg(0, 0);

     CommandQueue Q(context, devices[0], CL_QUEUE_PROFILING_ENABLE);
     Event ev;

     printf("The OpenCL runtime will lazily load the device program upon the \n");
     printf(" first enqueue of a kernel from the program, so the elapsed time \n");
     printf(" overall from the first enqueue will be longer to account for the \n");
     printf(" loading of the kernel.  Also, subsequent enqueue's of a kernel will \n");
     printf(" also potentially benefit from a warm cache on the device.\n\n");

     clock_gettime(CLOCK_MONOTONIC, &tp_start);
     Q.enqueueNDRangeKernel(kernel, NullRange, NDRange(1), NDRange(1), NULL, &ev);
     ev.wait();
     clock_gettime(CLOCK_MONOTONIC, &tp_end);

     double elapsed = clock_diff (&tp_start, &tp_end);
     printf("Elapsed (with Load): %8.6f secs\n", elapsed);
     ocl_event_times(ev, "Null kernel");

     for (int i = 0; i < 5; ++i)
     {
         clock_gettime(CLOCK_MONOTONIC, &tp_start);
         Q.enqueueNDRangeKernel(kernel, NullRange, NDRange(1), NDRange(1), NULL, &ev);
         ev.wait();
         clock_gettime(CLOCK_MONOTONIC, &tp_end);

         elapsed = clock_diff (&tp_start, &tp_end);
         printf("Elapsed (w/o  Load): %8.6f secs\n", elapsed);
         ocl_event_times(ev, "Null kernel");
     }
   }
   catch (Error err) 
   { cerr << "ERROR: " << err.what() << "(" << err.err() << ")" << endl; }

   cout << "Done!" << endl; 
}
